(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

(* The monadic combinators: return and bind. *)

(* The encoding of these combinators is independent of the encoding of the
   store. *)

(* ------------------------------------------------------------------------- *)

(* Return. *)

term return : box (univ (\a. a `arrow` monad a)) nil =
  /\ x a.
  \v : box a x.
  /\ y.
  \s : store (x `o` y).
  pack
    nil, (forward [a] [x] [y] v, s)
  as outcome a (x `o` y)

(* ------------------------------------------------------------------------- *)

(* Bind. *)

term bind :
  box (univ (\a. univ (\b. 
    (monad a `pair` (a `arrow` monad b)) `arrow` monad b
  ))) nil
  =
  /\ x a b.
  \fg : box (pair (monad a) (arrow a (monad b))) x.
  let (f, g) = box_pair [monad a] [arrow a (monad b)] [x] fg in
  assert f : box (monad a) x in
  assert g : box (arrow a (monad b)) x in
  /\ y.
  type xy = x `o` y in
  \s : store xy.
  let zxs : outcome a xy = f [y] s in
  unpack z, xs = zxs in
  let (x, s) = xs in
  type xyz = xy `o` z in
  assert x : box a xyz in
  assert s : store xyz in
  let outcome : outcome b xyz = g [y `o` z] x [nil] s in
  backward [b] [xy] [z] outcome

