Set Implicit Arguments.
Require Import DeBruijn.

Inductive tterm :=
| TVar: nat -> tterm
| TAbs: tterm -> tterm
| TApp: tterm -> tterm -> tterm
| TUnit
| TPair: tterm -> tterm -> tterm
| TLetPair: tterm -> tterm -> tterm.

Notation TLet t1 t2 :=
  (TApp (TAbs t2) t1).

Fixpoint lift_tterm (k : nat) (t : tterm) : tterm :=
  match t with
  | TVar x =>
      TVar (lift_var k x)
  | TAbs t =>
      TAbs (lift_tterm (1 + k) t)
  | TApp t1 t2 =>
      TApp (lift_tterm k t1) (lift_tterm k t2)
  | TUnit =>
      TUnit
  | TPair t1 t2 =>
      TPair (lift_tterm k t1) (lift_tterm k t2)
  | TLetPair t1 t2 =>
      TLetPair (lift_tterm k t1) (lift_tterm (2 + k) t2)
  end.

Fixpoint subst_tterm (u : tterm) (k : nat) (t : tterm) : tterm :=
  match t with
  | TVar x =>
      subst_var TVar u k x
  | TAbs t =>
      TAbs (subst_tterm (lift_tterm 0 u) (1 + k) t)
  | TApp t1 t2 =>
      TApp (subst_tterm u k t1) (subst_tterm u k t2)
  | TUnit =>
      TUnit
  | TPair t1 t2 =>
      TPair (subst_tterm u k t1) (subst_tterm u k t2)
  | TLetPair t1 t2 =>
      TLetPair (subst_tterm u k t1) (subst_tterm (lift_tterm 0 (lift_tterm 0 u)) (2 + k) t2)
  end.

Lemma subst_lift_tterm:
  subst_lift lift_tterm subst_tterm.
Proof.
  subst_lift_tac.
Qed.

Local Hint Resolve subst_lift_tterm.

Lemma lift_lift_tterm:
  lift_lift lift_tterm.
Proof.
  unfold lift_lift; intro a; induction a; intros; simpl in *; f_equal; eauto with lift_lift_db.
Qed.

Local Hint Resolve lift_lift_tterm.

Lemma lift_zero_lift_tterm:
  lift_zero_lift lift_tterm.
Proof.
  lift_zero_lift_tac.
Qed.

Ltac lzl_tterm :=
  lzl lift_tterm lift_zero_lift_tterm.

Hint Extern 1 => lzl_tterm : lift_subst_1_db.
Hint Extern 1 => lzl_tterm : lift_subst_2_db.
Hint Extern 1 => lzl_tterm : subst_subst_db.

Lemma lift_subst_1_tterm:
  lift_subst_1 lift_tterm lift_tterm subst_tterm.
Proof.
  lift_subst_1_tac.
Qed.

Local Hint Resolve lift_subst_1_tterm.

Hint Extern 1 => rewrite lift_subst_1_tterm; [ idtac | omega ] : subst_subst_db.

Lemma lift_subst_2_tterm:
  lift_subst_2 lift_tterm lift_tterm subst_tterm.
Proof.
  lift_subst_2_tac.
Qed.

Local Hint Resolve lift_subst_2_tterm.

Lemma subst_subst_tterm:
  subst_subst lift_tterm subst_tterm subst_tterm.
Proof.
  subst_subst_tac.
Qed.

Local Hint Resolve subst_subst_tterm.

Lemma pun_1_tterm:
  pun_1 TVar lift_tterm subst_tterm.
Proof.
  pun_tac.
Qed.

Lemma pun_2_tterm:
  pun_2 TVar lift_tterm subst_tterm.
Proof.
  pun_tac.
Qed.

Lemma lift_tterm_injective:
  lift_injective lift_tterm.
Proof.
  lift_injective_tac.
Qed.

Definition closed_tterm :=
  closed lift_tterm.

Lemma lift_closed_tterm:
  forall k t,
  closed_tterm t ->
  lift_tterm k t = t.
Proof.
  eauto using lift_closed.
Qed.

Lemma subst_closed_tterm:
  forall u k t,
  closed_tterm t ->
  subst_tterm u k t = t.
Proof.
  eauto using subst_closed.
Qed.

