(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Atom
open Kinds

(* The types in FORK are pure lambda-terms with constants. *)

type constant =
  | TyUnit
  | TyArrow
  | TyPair
  | TyForall of kind
  | TyExists of kind

type ty =
  | TyVar of atom
  | TyConstant of constant
  | TyAbs of atom * kind * locty
  | TyApp of locty * locty
  | TyConstraint of locty * kind

and locty =
    Error.location * ty

(* Types can be (recursively) defined by the user. A type signature
   maps atoms to types. *)

type signature =
    (kind * locty) AtomMap.t

(* A kind environment maps type variables to kinds. *)

type env =
    kind AtomMap.t

(* The kinds of the type constants. *)

val kind_of_constant: constant -> kind

(* Kind checking. *)

val infer: env -> locty -> kind
val check: signature -> env

